import json
import re

from mopidy_youtube import logger
from mopidy_youtube.apis.youtube_scrapi import scrAPI


# JSON based scrAPI
class jAPI(scrAPI):

    # search for videos and playlists using japi
    # **currently not working**
    @classmethod
    def run_search(cls, query):

        cls.session.headers = {
            "user-agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:66.0)"
            " Gecko/20100101 Firefox/66.0",
            "Cookie": "PREF=hl=en;",
            "Accept-Language": "en;q=0.5",
            "content_type": "application/json",
        }
        logger.info("session.get triggered: jAPI search")
        result = cls.session.get(cls.endpoint + "results", params=query)
        json_regex = r'window\["ytInitialData"] = ({.*?});'
        extracted_json = re.search(json_regex, result.text).group(1)
        result_json = json.loads(extracted_json)["contents"][
            "twoColumnSearchResultsRenderer"
        ]["primaryContents"]["sectionListRenderer"]["contents"][0][
            "itemSectionRenderer"
        ][
            "contents"
        ]
        return cls.json_to_items(cls, result_json)

    def json_to_items(cls, result_json):
        items = []
        for content in result_json:
            if "videoRenderer" in content:
                base = "videoRenderer"
            elif "compactVideoRenderer" in content:
                base = "compactVideoRenderer"
            elif "playlistVideoRenderer" in content:
                base = "playlistVideoRenderer"
            else:
                base = ""

            if base in [
                "videoRenderer",
                "compactVideoRenderer",
                "playlistVideoRenderer",
            ]:
                if "longBylineText" in content[base]:
                    byline = "longBylineText"
                else:
                    byline = "shortBylineText"

                try:
                    videoId = content[base]["videoId"]
                    logger.debug(videoId)
                except Exception as e:
                    # videoID = "Unknown"
                    logger.error("videoId exception %s" % e)
                    continue

                try:
                    title = content[base]["title"]["simpleText"]
                    logger.debug(title)
                except Exception:
                    try:
                        title = content[base]["title"]["runs"][0]["text"]
                        logger.debug(title)
                    except Exception as e:
                        # title = "Unknown"
                        logger.error("title exception %s" % e)
                        continue
                try:
                    channelTitle = content[base][byline]["runs"][0]["text"]
                    logger.debug(channelTitle)
                except Exception as e:
                    # channelTitle = "Unknown"
                    logger.error("channelTitle exception %s, %s" % (e, title))
                    continue

                item = {
                    "id": {"kind": "youtube#video", "videoId": videoId},
                    "snippet": {
                        "title": title,
                        "resourceId": {"videoId": videoId},
                        # TODO: full support for thumbnails
                        "thumbnails": {
                            "default": {
                                "url": "https://i.ytimg.com/vi/"
                                + videoId
                                + "/default.jpg",
                                "width": 120,
                                "height": 90,
                            },
                        },
                        "channelTitle": channelTitle,
                    },
                }

                try:
                    duration_text = content[base]["lengthText"]["simpleText"]
                    duration = "PT" + cls.format_duration(
                        re.match(cls.time_regex, duration_text)
                    )
                    logger.debug("duration: ", duration)
                except Exception as e:
                    logger.warn("no video-time, possibly live: ", e)
                    duration = "PT0S"

                item.update({"contentDetails": {"duration": duration}})

                # is channelId useful for anything?
                try:
                    channelId = content[base][byline]["runs"][0][
                        "navigationEndpoint"
                    ]["browseEndpoint"]["browseId"]
                    logger.debug(channelId)
                    item["snippet"].update({"channelId": channelId})
                except Exception as e:
                    logger.error("channelId exception %s, %s" % (e, title))

                items.append(item)

            elif "radioRenderer" in content:
                continue

            elif "playlistRenderer" in content:
                item = {
                    "id": {
                        "kind": "youtube#playlist",
                        "playlistId": content["playlistRenderer"]["playlistId"],
                    },
                    "contentDetails": {
                        "itemCount": content["playlistRenderer"]["videoCount"]
                    },
                    "snippet": {
                        "title": content["playlistRenderer"]["title"][
                            "simpleText"
                        ],
                        # TODO: full support for thumbnails
                        "thumbnails": {
                            "default": {
                                "url": "https://i.ytimg.com/vi/"
                                + content["playlistRenderer"][
                                    "navigationEndpoint"
                                ]["watchEndpoint"]["videoId"]
                                + "/default.jpg",
                                "width": 120,
                                "height": 90,
                            },
                        },
                        "channelTitle": content["playlistRenderer"][
                            "longBylineText"
                        ]["runs"][0]["text"],
                    },
                }

                items.append(item)

        # remove duplicates
        items[:] = [
            json.loads(t)
            # for t in {json.dumps(d) for d in items}
            for t in {json.dumps(d, sort_keys=True) for d in items}
        ]
        return items



import pykka
from mopidy.core import listener
from mopidy_youtube import backend, logger, youtube

autoplay_enabled = False
strict_autoplay = False
max_autoplay_length = 600
autoplayed = []
max_degrees_of_separation = 3


class YouTubeAutoplayer(pykka.ThreadingActor, listener.CoreListener):
    def __init__(self, config, core):
        super().__init__()
        self.config = config
        self.core = core
        self.autoplay_enabled = config["youtube"]["autoplay_enabled"]
        self.strict_autoplay = config["youtube"]["strict_autoplay"]
        self.max_degrees_of_separation = config["youtube"][
            "max_degrees_of_separation"
        ]
        self.max_autoplay_length = config["youtube"]["max_autoplay_length"]
        self.base_track_id = ""
        self.degrees_of_separation = 0

    # Called by mopidy on start of playback of a URI
    # This function emulates the youtube autoplay functionality by retrieving the most
    # most related video to a video just played by a youtube API call, adding this new
    # video URI to the tracklist
    #
    # With the option "strict_autoplay" enabled, the next played URI will be the newly
    # added video. Without the option "strict_autoplay" enabled [default], the autoplay
    # functionality will only be executed if the end of the current tracklist is reached
    #
    # The autoplay functionality will not work correctly in combination with the repeat
    # option and is therefore disabled if repeat is enabled

    def track_playback_started(self, tl_track):
        if not self.autoplay_enabled:
            return None

        [tlTrackId, track] = tl_track

        if "youtube:video/" not in track.uri:
            return None

        try:
            tl = self.core.tracklist

            if tl.get_repeat().get() is True:
                logger.info(
                    "Autoplayer: will not add tracks when repeat is enabled."
                )
                return None

            if tl.get_random().get() is True:
                logger.info(
                    "Autoplayer: shuffle will not work when autoplay is enabled."
                )

            if self.strict_autoplay is False:
                tlTracks = tl.get_tl_tracks().get()
                if len(tlTracks) != 0:
                    if tlTrackId is not tlTracks[-1].tlid:
                        logger.debug(
                            "Autoplayer: called not at end of track list."
                        )
                        return None
                    elif tl.get_consume().get() is True:
                        logger.warning(
                            "Autoplayer: when having consume track enabled, "
                            'try with "strict_autoplay" option enabled for '
                            "better results"
                        )
                        return None

            current_track_id = backend.extract_id(track.uri)

            if current_track_id not in autoplayed:
                self.base_track_id = current_track_id
                autoplayed.append(current_track_id)  # avoid replaying track
                logger.info("setting new autoplay base id")

            if self.max_degrees_of_separation:
                if self.degrees_of_separation < self.max_degrees_of_separation:
                    self.degrees_of_separation += 1
                    logger.info("incrementing autoplay degrees of separation")
                else:
                    current_track_id = self.base_track_id
                    self.degrees_of_separation = 0
                    logger.info("resetting autoplay base track id")

            related_videos = youtube.Video.related_videos(current_track_id)

            # remove already autoplayed
            related_videos[:] = [
                related_video
                for related_video in related_videos
                if related_video.id not in autoplayed
            ]
            # remove if track_length is 0 (probably a live video) or None
            related_videos[:] = [
                related_video
                for related_video in related_videos
                if related_video.length.get()
            ]
            # remove if too long
            if self.max_autoplay_length:
                related_videos[:] = [
                    related_video
                    for related_video in related_videos
                    if related_video.length.get() < self.max_autoplay_length
                ]

            if len(related_videos) == 0:
                logger.info("could not get related videos: ending autoplay")
                return None
            else:
                next_video = related_videos[0]
                autoplayed.append(next_video.id)
                name = next_video.title.get()
                uri = [
                    "youtube:video/%s.%s"
                    % (backend.safe_url(name), next_video.id)
                ]
                tl.add(uris=uri).get()
                return None

        except Exception as e:
            logger.error('Autoplayer error "%s"', e)
            return None



import json
import re
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from mopidy_youtube import logger
from mopidy_youtube.apis.youtube_japi import jAPI
from mopidy_youtube.apis.youtube_scrapi import scrAPI


class bs4API(scrAPI):
    """
    Indirect access to YouTube data, without API
    using BS4 (instead of regex, as used by scrAPI)
    """

    time_regex = (
        r"(?:(?:(?P<durationHours>[0-9]+)\:)?"
        r"(?P<durationMinutes>[0-9]+)\:"
        r"(?P<durationSeconds>[0-9]{2}))"
    )

    @classmethod
    def run_search(cls, query):

        result = cls.session.get(urljoin(cls.endpoint, "results"), params=query)
        if result.status_code == 200:
            soup = BeautifulSoup(result.text, "html.parser")
            # hack because youtube result sometimes seem to be missing a </script> tag
            if not soup.find_all(
                "div", class_=["yt-lockup-video", "yt-lockup-playlist"]
            ):
                for script in soup.find_all("script"):
                    # assume scripts should not contain "<div>" tags
                    # and that, if they do, there is a missing "</script>" tag
                    if script.find(text=re.compile("<div")):
                        new_soup = BeautifulSoup(
                            re.sub(
                                "<div", "</script><div", script.string, count=1
                            ),
                            "html.parser",
                        )
                        script.replace_with(new_soup)

            # strip out ads that are returned
            [
                ad.decompose()
                for ad in soup.find_all("div", class_="pyv-afc-ads-container")
            ]

            results = soup.find_all(
                "div", class_=["yt-lockup-video", "yt-lockup-playlist"]
            )

            if not results:
                logger.info("nothing in the soup, trying japi")
                json_regex = r'window\["ytInitialData"] = ({.*?});'
                extracted_json = re.search(json_regex, result.text).group(1)
                result_json = json.loads(extracted_json)["contents"][
                    "twoColumnSearchResultsRenderer"
                ]["primaryContents"]["sectionListRenderer"]["contents"][0][
                    "itemSectionRenderer"
                ][
                    "contents"
                ]
                return jAPI.json_to_items(cls, result_json)
            else:
                return cls.soup_to_items(cls, results)

    def soup_to_items(cls, results):
        items = []
        for result in results:
            if "yt-lockup-video" in result.get("class"):
                try:
                    videoId = result["data-context-item-id"]
                except Exception as e:
                    # videoID = "Unknown"
                    logger.error("videoId exception %s" % e)
                    continue

                try:
                    title_text = result.find("a", {"aria-describedby": True})
                    title = title_text.text.strip()
                except Exception:
                    try:
                        title_text = result.find("a", {"title": True})
                        title = title_text.text.strip()
                    except Exception as e:
                        # title = "Unknown"
                        logger.error("title exception %s" % e, videoId)
                        continue

                try:
                    channelTitle_text = result.find(class_="yt-lockup-byline")
                    channelTitle = channelTitle_text.text.strip()
                except Exception as e:
                    # channelTitle = "Unknown"
                    logger.error("channelTitle exception %s" % e)
                    continue

                item = {
                    "id": {"kind": "youtube#video", "videoId": videoId},
                    "snippet": {
                        "title": title,
                        # TODO: full support for thumbnails
                        "thumbnails": {
                            "default": {
                                "url": "https://i.ytimg.com/vi/"
                                + videoId
                                + "/default.jpg",
                                "width": 120,
                                "height": 90,
                            },
                        },
                        "channelTitle": channelTitle,
                    },
                }
                try:
                    duration_text = result.find(class_="video-time").text
                    duration = "PT" + cls.format_duration(
                        re.match(cls.time_regex, duration_text)
                    )
                except Exception as e:
                    logger.info(
                        "no video-time, possibly live: ",
                        e,
                        result["data-context-item-id"],
                    )
                    duration = "PT0S"
                item.update({"contentDetails": {"duration": duration}})
                items.append(item)

            elif "yt-lockup-playlist" in result.get("class"):
                item = {
                    "id": {
                        "kind": "youtube#playlist",
                        "playlistId": result.find(class_="yt-lockup-title")
                        .next["href"]
                        .partition("list=")[2],
                    },
                    "contentDetails": {
                        "itemCount": re.sub(
                            r"[^\d\.]",
                            "",
                            result.find(
                                class_="formatted-video-count-label"
                            ).text.split(" ")[0],
                        )
                    },
                    "snippet": {
                        "title": result.find(
                            class_="yt-lockup-title"
                        ).next.text,
                        # TODO: full support for thumbnails
                        "thumbnails": {
                            "default": {
                                "url": (
                                    "https://i.ytimg.com/vi/"
                                    + result.find(class_="yt-lockup-thumbnail")
                                    .find("a")["href"]
                                    .partition("v=")[2]
                                    .partition("&")[0]
                                    + "/default.jpg"
                                ),
                                "width": 120,
                                "height": 90,
                            },
                        },
                        "channelTitle": result.find(
                            class_="yt-lockup-byline"
                        ).text,
                    },
                }
                # don't append radiolist playlists
                if str(item["id"]["playlistId"]).startswith("PL"):
                    items.append(item)
        return items

    @classmethod
    def list_playlistitems(cls, id, page, max_results):
        query = {"list": id, "app": "desktop", "persist_app": 1}
        logger.info("session.get triggered: list_playlist_items")
        ajax_css = "button[data-uix-load-more-href]"

        videos = items = []
        if page == "":
            result = cls.session.get(
                urljoin(cls.endpoint, "playlist"), params=query
            )
            if result.status_code == 200:
                soup = BeautifulSoup(result.text, "html.parser")
        else:
            result = cls.session.get(urljoin(cls.endpoint, page))
            if result.status_code == 200:
                soup = BeautifulSoup(
                    "".join(result.json().values()), "html.parser"
                )
        if soup:
            # get load more button
            full_ajax = soup.select(ajax_css)
            if len(full_ajax) > 0:
                ajax = full_ajax[0]["data-uix-load-more-href"]
            else:
                ajax = None
            videos = [
                video
                for video in soup.find_all("tr", {"class": "pl-video"})
                if all(
                    [
                        video.find(class_="timestamp"),
                        video.find(class_="pl-video-owner"),
                    ]
                )
            ]

            if not videos:
                logger.info(
                    "nothing in the soup, trying japi list_playlistitems"
                )
                json_regex = r'window\["ytInitialData"] = ({.*?});'
                extracted_json = re.search(json_regex, result.text).group(1)
                result_json = json.loads(extracted_json)["contents"][
                    "twoColumnBrowseResultsRenderer"
                ]["tabs"][0]["tabRenderer"]["content"]["sectionListRenderer"][
                    "contents"
                ][
                    0
                ][
                    "itemSectionRenderer"
                ][
                    "contents"
                ][
                    0
                ][
                    "playlistVideoListRenderer"
                ][
                    "contents"
                ]
                items = jAPI.json_to_items(cls, result_json)
            else:
                items = cls.plsoup_to_items(cls, videos)

            return json.loads(
                json.dumps(
                    {"nextPageToken": ajax, "items": items},
                    sort_keys=False,
                    indent=1,
                )
            )

    def plsoup_to_items(cls, videos):
        items = []
        for video in videos:
            item = {
                "id": video["data-video-id"],
                "contentDetails": {
                    "duration": "PT"
                    + cls.format_duration(
                        re.match(
                            cls.time_regex, video.find(class_="timestamp").text,
                        )
                    ),
                },
                "snippet": {
                    "resourceId": {"videoId": video["data-video-id"]},
                    "title": video["data-title"],
                    # TODO: full support for thumbnails
                    "thumbnails": {
                        "default": {
                            "url": "https://i.ytimg.com/vi/"
                            + video["data-video-id"]
                            + "/default.jpg",
                            "width": 120,
                            "height": 90,
                        },
                    },
                    "channelTitle": video.find(class_="pl-video-owner")
                    .find("a")
                    .text,
                },
            }

            items.append(item)

        return items

    @classmethod
    def list_videos(cls, ids):
        """
        list videos - EXPERIMENTAL, using exact search for ids
        """
        logger.info("session.get triggered: list_videos (experimental)")

        items = []

        rs = [
            {
                "search_query": '"' + id + '"',
                "sp": "EgIQAQ%3D%3D",
                "app": "desktop",
                "persist_app": 1,
            }
            for id in ids
        ]
        results = [
            result
            for r in rs
            for result in cls.run_search(r)
            if result["id"]["videoId"] in ids
        ]
        for result in results:
            result.update({"id": result["id"]["videoId"]})
            items.extend([result])
        return json.loads(
            json.dumps({"items": items}, sort_keys=False, indent=1)
        )

    @classmethod
    def list_playlists(cls, ids):
        """
        list playlists - EXPERIMENTAL, using exact search for ids
        """
        logger.info("session.get triggered: list_playlists (experimental)")

        items = []

        rs = [
            {
                "search_query": '"' + id + '"',
                "sp": "EgIQAw%3D%3D",
                "app": "desktop",
                "persist_app": 1,
            }
            for id in ids
        ]
        results = [
            result
            for r in rs
            for result in cls.run_search(r)
            if result["id"]["playlistId"] in ids
        ]
        for result in results:
            result.update({"id": result["id"]["playlistId"]})
            items.extend([result])

        return json.loads(
            json.dumps({"items": items}, sort_keys=False, indent=1)
        )

    @classmethod
    def list_related_videos(cls, video_id):
        """
        returns related videos for a given video_id
        """
        items = []

        query = {"v": video_id, "app": "desktop", "persist_app": 1}
        logger.info("session.get triggered: list_related_videos")
        result = cls.session.get(cls.endpoint + "watch", params=query)
        if result.status_code == 200:
            soup = BeautifulSoup(result.text, "html.parser")
            results = soup.find_all("li", class_=["related-list-item"])
            if not results:
                logger.info("nothing in the soup, trying japi related videos")
                json_regex = r'window\["ytInitialData"] = ({.*?});'
                extracted_json = re.search(json_regex, result.text).group(1)
                result_json = json.loads(extracted_json)["contents"][
                    "twoColumnWatchNextResults"
                ]["secondaryResults"]["secondaryResults"][
                    "results"
                ]  # noqa: E501
                items = jAPI.json_to_items(cls, result_json)

            for result in results:
                if "related-list-item-compact-video" in result.get("class"):

                    try:
                        videoId = result.find("span", {"data-vid": True})[
                            "data-vid"
                        ]
                    except Exception as e:
                        # videoID = "Unknown"
                        logger.error("videoId exception %s" % e)
                        continue

                    try:
                        title_text = result.find("span", class_="title")
                        title = title_text.text.strip()
                    except Exception as e:
                        # title = "Unknown"
                        logger.error("title exception %s" % e)
                        continue

                    try:
                        channelTitle_text = result.find(
                            "span", class_="stat attribution"
                        )
                        channelTitle = channelTitle_text.text.strip()
                    except Exception as e:
                        # channelTitle = "Unknown"
                        logger.error("channelTitle exception %s" % e)
                        continue

                    item = {
                        "id": {"kind": "youtube#video", "videoId": videoId},
                        "snippet": {
                            "title": title,
                            "resourceId": {"videoId": videoId},
                            # TODO: full support for thumbnails
                            "thumbnails": {
                                "default": {
                                    "url": "https://i.ytimg.com/vi/"
                                    + videoId
                                    + "/default.jpg",
                                    "width": 120,
                                    "height": 90,
                                },
                            },
                            "channelTitle": channelTitle,
                        },
                    }

                    try:
                        duration_text = result.find(class_="video-time")
                        duration = "PT" + cls.format_duration(
                            re.match(cls.time_regex, duration_text.text)
                        )
                        item.update({"contentDetails": {"duration": duration}})
                    except Exception as e:
                        # don't set duration on exception
                        logger.warn("duration exception %s" % e)
                        # because duration error doesn't invalidate item (live tracks, eg)
                    items.append(item)

        return json.loads(
            json.dumps({"items": items}, sort_keys=False, indent=1)
        )



import logging
import pathlib

import pkg_resources

from mopidy import config, ext

__version__ = pkg_resources.get_distribution("Mopidy-YouTube").version

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = "Mopidy-YouTube"
    ext_name = "youtube"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["youtube_api_key"] = config.String(optional=True)
        schema["threads_max"] = config.Integer(minimum=1)
        schema["search_results"] = config.Integer(minimum=1)
        schema["playlist_max_videos"] = config.Integer(minimum=1)
        schema["api_enabled"] = config.Boolean()
        schema["autoplay_enabled"] = config.Boolean(optional=True)
        schema["strict_autoplay"] = config.Boolean(optional=True)
        schema["max_autoplay_length"] = config.Integer()
        schema["max_degrees_of_separation"] = config.Integer()
        return schema

    def setup(self, registry):
        from .backend import YouTubeBackend
        from .frontend import YouTubeAutoplayer

        registry.add("backend", YouTubeBackend)
        registry.add("frontend", YouTubeAutoplayer)



import re
import string
import unicodedata
from urllib.parse import parse_qs, urlparse

import pykka
from mopidy import backend, httpclient
from mopidy.models import Album, Artist, SearchResult, Track
from mopidy_youtube import Extension, logger, youtube
from mopidy_youtube.apis import youtube_api, youtube_bs4api


"""
A typical interaction:
1. User searches for a keyword (YouTubeLibraryProvider.search)
2. User adds a track to the queue (YouTubeLibraryProvider.lookup)
3. User plays a track from the queue (YouTubePlaybackProvider.translate_uri)
step 1 requires only 2 API calls. Data for the next steps are loaded in the
background, so steps 2/3 are usually instantaneous.
"""


def extract_id(uri):
    """takes a uri in the form youtube:video/<title>.<id> and returns <id>"""
    return uri.split(".")[-1]


def safe_url(uri):
    valid_chars = f"-_.() {string.ascii_letters}{string.digits}"
    safe_uri = unicodedata.normalize("NFKD", uri).encode("ASCII", "ignore")
    return re.sub(
        r"\s+", " ", "".join(c for c in map(chr, safe_uri) if c in valid_chars)
    ).strip()


class YouTubeBackend(pykka.ThreadingActor, backend.Backend):
    def __init__(self, config, audio):
        super().__init__()
        self.config = config
        self.library = YouTubeLibraryProvider(backend=self)
        self.playback = YouTubePlaybackProvider(audio=audio, backend=self)
        youtube_api.youtube_api_key = (
            config["youtube"]["youtube_api_key"] or None
        )
        youtube.ThreadPool.threads_max = config["youtube"]["threads_max"]
        youtube.Video.search_results = config["youtube"]["search_results"]
        youtube.Playlist.playlist_max_videos = config["youtube"][
            "playlist_max_videos"
        ]
        youtube.api_enabled = config["youtube"]["api_enabled"]
        self.uri_schemes = ["youtube", "yt"]
        self.user_agent = "{}/{}".format(Extension.dist_name, Extension.version)

    def on_start(self):
        proxy = httpclient.format_proxy(self.config["proxy"])
        youtube.Video.proxy = proxy
        headers = {
            "user-agent": httpclient.format_user_agent(self.user_agent),
            "Cookie": "PREF=hl=en;",
            "Accept-Language": "en;q=0.8",
        }

        if youtube.api_enabled is True:
            if youtube_api.youtube_api_key is None:
                logger.error("No YouTube API key provided, disabling API")
                youtube.api_enabled = False
            else:
                youtube.Entry.api = youtube_api.API(proxy, headers)
                if youtube.Entry.search(q="test") is None:
                    logger.error(
                        "Failed to verify YouTube API key, disabling API"
                    )
                    youtube.api_enabled = False
                else:
                    logger.info("YouTube API key verified")

        if youtube.api_enabled is False:
            logger.info("using bs4API")
            youtube.Entry.api = youtube_bs4api.bs4API(proxy, headers)


class YouTubeLibraryProvider(backend.LibraryProvider):
    """
    Called when browsing or searching the library. To avoid horrible browsing
    performance, and since only search makes sense for youtube anyway, we we
    only answer queries for the 'any' field (for instance a {'artist': 'U2'}
    query is ignored).

    For performance we only do 2 API calls before we reply, one for search
    (youtube.Entry.search) and one to fetch video_count of all playlists
    (youtube.Playlist.load_info).

    We also start loading 2 things in the background:
     - info for all videos
     - video list for all playlists
    Hence, adding search results to the playing queue (see
    YouTubeLibraryProvider.lookup) will most likely be instantaneous, since
    all info will be ready by that time.
    """

    def search(self, query=None, uris=None, exact=False):
        # TODO Support exact search
        logger.info('youtube LibraryProvider.search "%s"', query)

        # handle only searching (queries with 'any') not browsing!
        if not (query and "any" in query):
            return None

        search_query = " ".join(query["any"])
        logger.info('Searching YouTube for query "%s"', search_query)

        try:
            entries = youtube.Entry.search(search_query)
        except Exception as e:
            logger.error('search error "%s"', e)
            return None

        # load playlist info (to get video_count) of all playlists together
        playlists = [entry for entry in entries if not entry.is_video]
        youtube.Playlist.load_info(playlists)

        albums = []
        artists = []
        tracks = []

        for entry in entries:
            if entry.is_video:
                uri_base = "youtube:video"
                album = "YouTube Video"
                length = int(entry.length.get()) * 1000

                # # does it make sense to try to return youtube 'channels' as
                # # mopidy 'artists'? I'm not convinced.
                # if entry.channelId.get():
                #     artists.append(
                #             Artist(
                #                 name=f"YouTube channel: {entry.channel.get()}",
                #                 uri="yotube:channel/%s.%s" % (safe_url(entry.channel.get()), entry.channelId.get()),
                #                 )
                #             )
                # else:
                #     logger.info("no channelId")

            else:
                uri_base = "youtube:playlist"
                album = "YouTube Playlist (%s videos)" % entry.video_count.get()
                length = 0

            name = entry.title.get()

            tracks.append(
                Track(
                    name=name.replace(";", ""),  # why is this .replace here?
                    comment=entry.id,
                    length=length,
                    artists=[Artist(name=entry.channel.get())],
                    album=Album(name=album),
                    uri="%s/%s.%s" % (uri_base, safe_url(name), entry.id),
                )
            )

        # load video info and playlist videos in the background. they should be
        # ready by the time the user adds search results to the playing queue
        for pl in playlists:
            name = pl.title.get()

            # does it make sense to return youtube 'playlists' as mopidy 'albums'?
            albums.append(
                Album(
                    name=name,
                    artists=[
                        Artist(
                            name=f"YouTube Playlist ({pl.video_count.get()} videos)"
                        )
                    ],
                    uri="%s/%s.%s"
                    % ("youtube:playlist", safe_url(name), pl.id),
                )
            )

            pl.videos  # start loading

        return SearchResult(
            uri="youtube:search", tracks=tracks, artists=artists, albums=albums
        )

    def lookup(self, uri):
        """
        Called when the user adds a track to the playing queue, either from the
        search results, or directly by adding a yt:http://youtube.com/.... uri.
        uri can be of the form
            [yt|youtube]:<url to youtube video>
            [yt|youtube]:<url to youtube playlist>
            youtube:video/<title>.<id>
            youtube:playlist/<title>.<id>

        If uri is a video then a single track is returned. If it's a playlist the
        list of all videos in the playlist is returned.

        We also start loading the audio_url of all videos in the background, to
        be ready for playback (see YouTubePlaybackProvider.translate_uri).
        """

        logger.info('youtube LibraryProvider.lookup "%s"', uri)

        video_id = None
        playlist_id = None

        # # should channels be returned?
        # channel_id = None
        if "youtube.com" in uri:
            url = urlparse(uri.replace("yt:", "").replace("youtube:", ""))
            req = parse_qs(url.query)
            if "list" in req:
                playlist_id = req.get("list")[0]
            else:
                video_id = req.get("v")[0]

        elif "youtu.be" in uri:
            url = uri.replace("yt:", "").replace("youtube:", "")
            if not re.match("(?:http|https)://", url):
                url = "http://" + url
            video_id = urlparse(url).path
            if video_id[0] == "/":
                video_id = video_id[1:]

        elif "video/" in uri:
            video_id = extract_id(uri)
        # elif "channel/" in uri:
        #     channel_id = extract_id(uri)
        else:
            playlist_id = extract_id(uri)

        if video_id:
            video = youtube.Video.get(video_id)
            video.audio_url  # start loading
            video.title.get()
            return [
                Track(
                    name=video.title.get().replace(
                        ";", ""
                    ),  # why is this .replace here?
                    comment=video.id,
                    length=video.length.get() * 1000,
                    artists=[Artist(name=video.channel.get())],
                    album=Album(name="YouTube Video",),
                    uri="youtube:video/%s.%s"
                    % (safe_url(video.title.get()), video.id),
                )
            ]

        # elif channel_id:
        #     logger.info(channel_id)
        #     channel = youtube.Channel.get(channel_id)
        #
        #     if not channel.videos.get():
        #         logger.info('Cannot load "%s"', uri)
        #         return []
        #     videos = [v for v in channel.videos.get() if v.length.get() is not None]
        #     album_name = "YouTube Video"

        else:
            playlist = youtube.Playlist.get(playlist_id)

            if not playlist.videos.get():
                logger.error('Cannot load "%s"', uri)
                return []

            # ignore videos for which no info was found (removed, etc)
            videos = [
                v for v in playlist.videos.get() if v.length.get() is not None
            ]
            album_name = playlist.title.get()

        # load audio_url in the background to be ready for playback
        for video in videos:
            video.audio_url  # start loading

        return [
            Track(
                name=video.title.get().replace(
                    ";", ""
                ),  # why is this .replace here?
                comment=video.id,
                length=video.length.get() * 1000,
                track_no=count,
                artists=[Artist(name=video.channel.get())],
                album=Album(name=album_name),
                uri="youtube:video/%s.%s"
                % (safe_url(video.title.get()), video.id),
            )
            for count, video in enumerate(videos, 1)
        ]

    def get_images(self, uris):
        return {uri: youtube.Video.get(uri).thumbnails.get() for uri in uris}


class YouTubePlaybackProvider(backend.PlaybackProvider):
    def translate_uri(self, uri):
        """
        Called when a track us ready to play, we need to return the actual url of
        the audio. uri must be of the form youtube:video/<title>.<id>
        (only videos can be played, playlists are expended into tracks by
        YouTubeLibraryProvider.lookup)
        """

        logger.info('youtube PlaybackProvider.translate_uri "%s"', uri)

        if "youtube:video/" not in uri:
            return None

        try:
            id = extract_id(uri)
            return youtube.Video.get(id).audio_url.get()
        except Exception as e:
            logger.error('translate_uri error "%s"', e)
            return None



import youtube_dl
from mopidy import httpclient

# import os
from unittest import mock
import pytest
import vcr
from mopidy_youtube import Extension, backend, youtube
from mopidy_youtube.apis import youtube_api, youtube_bs4api, youtube_scrapi

proxy = None  # httpclient.format_proxy(config['proxy'])
youtube.Video.proxy = proxy

user_agent = "{}/{}".format(Extension.dist_name, Extension.version)

headers = {
    "user-agent": httpclient.format_user_agent(user_agent),
    "Cookie": "PREF=hl=en;",
    "Accept-Language": "en;q=0.8",
}

# youtube_api.youtube_api_key = os.environ.get("YOUTUBE_API_KEY", "FAKE_API_KEY")

yak = "fakeyoutubekey"
youtube_api.youtube_api_key = yak


@pytest.yield_fixture
def youtube_dl_mock():
    patcher = mock.patch.object(youtube, "youtube_dl", spec=youtube_dl)
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def youtube_dl_mock_with_video(youtube_dl_mock):
    video_mock = youtube_dl_mock.YoutubeDL.return_value
    video_mock.bigthumb = "big thumb"
    video_mock.bigthumbhd = "big thumb in hd"
    video_mock.getbestaudio.return_value.url = "http://example.com/"
    video_mock.extract_info.return_value.url = "http://example.com/"
    video_mock.length = 2000
    video_mock.title = "a title"
    video_mock.videoid = "a video id"
    return youtube_dl_mock


@pytest.fixture
def config():
    return {
        "core": {"cache_dir": "."},
        "http": {"hostname": "127.0.0.1", "port": "6680"},
        "youtube": {
            "enabled": "true",
            "youtube_api_key": yak,
            "threads_max": 16,
            "search_results": 30,
            "playlist_max_videos": 20,
            "api_enabled": False,
        },
    }


def get_backend(config):
    return backend.YouTubeBackend(config=config, audio=mock.Mock())


def test_uri_schemes(config):
    backend_inst = get_backend(config)

    assert "youtube" in backend_inst.uri_schemes
    assert "yt" in backend_inst.uri_schemes


def test_init_sets_up_the_providers(config):
    backend_inst = get_backend(config)

    assert isinstance(backend_inst.library, backend.YouTubeLibraryProvider)
    assert isinstance(backend_inst.playback, backend.YouTubePlaybackProvider)


# need to work out how to add youtube_api.API to this list
apis = [youtube_scrapi.scrAPI, youtube_bs4api.bs4API]


@vcr.use_cassette(
    "tests/fixtures/youtube_playlist.yaml", filter_query_parameters=["key"]
)
def test_get_playlist(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        pl = youtube.Playlist.get("PLvdVG7oER2eFutjd4xl3TGNDui9ELvY4D")

        assert len(pl.videos.get()) == 20
        assert pl.videos.get()[0].title.get()

        # Playlist.videos starts loading video info in the background
        video = pl.videos.get()[0]
        assert video._length  # should be ready
        assert video.length.get() == 392

        pl2 = youtube.Playlist.get("PLvdVG7oER2eFutjd4xl3TGNDui9ELvY4D")

        assert pl2 is pl  # fetch from cache
        assert pl._videos  # should be ready


@vcr.use_cassette(
    "tests/fixtures/youtube_list_playlists.yaml",
    filter_query_parameters=["key"],
)
def test_list_playlists(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        playlists = api.list_playlists(
            [
                "PLfGibfZATlGq6zF72No5BZaScBiWWb6U1",
                "PLRHAVCbqFwJBkRupIhuGW3_XEIWc-ZYER",
            ]
        )

        assert len(playlists["items"]) == 2


@vcr.use_cassette(
    "tests/fixtures/youtube_search.yaml", filter_query_parameters=["key"]
)
def test_search(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)
        backend_inst = get_backend(config)

        videos = backend_inst.library.search(query={"omit-any": ["chvrches"]})
        assert videos is None

        videos = backend_inst.library.search(query={"any": ["chvrches"]})
        assert len(videos.tracks) == 30

        videos = youtube.Entry.search("chvrches")
        assert len(videos) == 30
        assert videos[0]._title  # should be ready
        assert videos[0]._channel  # should be ready
        assert videos[0]._length  # should be ready (scrAPI)

        video = youtube.Video.get("BZyzX4c1vIs")

        assert video in videos  # cached


@vcr.use_cassette(
    "tests/fixtures/youtube_lookup.yaml", filter_query_parameters=["key"]
)
def test_lookup(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)
        backend_inst = get_backend(config)

        video_uris = [
            "yt:https://www.youtube.com/watch?v=s5BJXwNeKsQ",
            "youtube:https://www.youtube.com/watch?v=1lWJXDG2i0A",
            "youtube:video/Tom Petty And The Heartbreakers - "
            "I Won't Back Down (Official Music Video).nvlTJrNJ5lA",
        ]
        for video_uri in video_uris:
            video = backend_inst.library.lookup(video_uri)
            assert len(video) == 1

        playlist_uris = [
            "yt:https://www.youtube.com/watch?v=SIhb-kNvL6M"
            "&list=PLo4c-riVwz2miWOT3Y2VWzg2bmV4FmC8J",
            "youtube:https://www.youtube.com/watch?v=lis8WGZQ9tw"
            "&list=PLW3M-yio9tLtQLihn1wrJYzuV7AUPMq63",
            "youtube:playlist/Tom Petty and The Heartbreakers GREATEST HITS "
            "(Complete Album).PLrpyDacBCh7Bs3cNzKtMbuafcK_nw9znk",
        ]
        for playlist_uri in playlist_uris:
            playlist = backend_inst.library.lookup(playlist_uri)
            assert len(playlist) == 18


@vcr.use_cassette(
    "tests/fixtures/youtube_get_video.yaml", filter_query_parameters=["key"]
)
def test_get_video(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        video = youtube.Video.get("e1YqueG2gtQ")

        assert video.length.get()

        # get again, should fetch from cache, _length should be ready
        video2 = youtube.Video.get("e1YqueG2gtQ")

        assert video2 is video
        assert video2._length


@vcr.use_cassette(
    "tests/fixtures/youtube_list_videos.yaml", filter_query_parameters=["key"]
)
def test_list_videos(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        videos = api.list_videos(["_mTRvJ9fugM", "h_uyq8oGDvU"])

        assert len(videos["items"]) == 2


@vcr.use_cassette(
    "tests/fixtures/youtube_list_playlistitems.yaml",
    filter_query_parameters=["key"],
)
def test_list_playlistitems(config):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        playlistitems = api.list_playlistitems(
            "PLHZ4bmFPEGndpOL8j_Xdgu7ainO7mf7nY", None, 20
        )

        assert len(playlistitems["items"]) == 20


def test_audio_url(youtube_dl_mock_with_video):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        video = youtube.Video.get("e1YqueG2gtQ")

        assert video.audio_url.get()


def test_audio_url_fail(youtube_dl_mock):

    for api in apis:
        youtube.Entry.api = api(proxy, headers)

        youtube_dl_mock.YoutubeDL.side_effect = Exception("Removed")

        video = youtube.Video.get("unknown")

        assert not video.audio_url.get()



from mopidy_youtube import logger
from mopidy_youtube.youtube import Client, Video

youtube_api_key = None


class API(Client):
    """
    Direct access to YouTube Data API
    see https://developers.google.com/youtube/v3/docs/
    """

    endpoint = "https://www.googleapis.com/youtube/v3/"

    @classmethod
    def search(cls, q):
        """
        search for both videos and playlists using a single API call
        see https://developers.google.com/youtube/v3/docs/search
        """

        query = {
            "part": "id, snippet",
            "fields": "items(id, snippet(title, thumbnails(default), channelTitle))",
            "maxResults": Video.search_results,
            "type": "video,playlist",
            "q": q,
            "key": youtube_api_key,
        }
        logger.info("session.get triggered: search")
        result = cls.session.get(API.endpoint + "search", params=query)
        return result.json()

    @classmethod
    def list_related_videos(cls, video_id):
        """
        queries related videos to a given video_id using a single API call
        https://developers.google.com/youtube/v3/docs/search
        """

        query = {
            "relatedToVideoId": video_id,
            "part": "snippet",
            "maxResults": 20,
            "type": "video",
            "key": youtube_api_key,
        }
        logger.info("session.get triggered: list_related_videos")
        result = cls.session.get(API.endpoint + "search", params=query)
        return result.json()

    @classmethod
    def list_videos(cls, ids):
        """
        list videos
        see https://developers.google.com/youtube/v3/docs/videos/list
        """

        query = {
            "part": "id,snippet,contentDetails",
            "fields": "items(id,snippet(title,channelTitle),"
            + "contentDetails(duration))",
            "id": ",".join(ids),
            "key": youtube_api_key,
        }
        logger.info("session.get triggered: list_videos")
        result = cls.session.get(API.endpoint + "videos", params=query)
        return result.json()

    @classmethod
    def list_playlists(cls, ids):
        """
        list playlists
        see https://developers.google.com/youtube/v3/docs/playlists/list
        """

        query = {
            "part": "id,snippet,contentDetails",
            "fields": "items(id,snippet(title,thumbnails,channelTitle),"
            + "contentDetails(itemCount))",
            "id": ",".join(ids),
            "key": youtube_api_key,
        }
        logger.info("session.get triggered: list_playlists")
        result = cls.session.get(API.endpoint + "playlists", params=query)
        return result.json()

    @classmethod
    def list_playlistitems(cls, id, page, max_results):
        """
        list playlist items
        see https://developers.google.com/youtube/v3/docs/playlistItems/list
        """

        query = {
            "part": "id,snippet",
            "fields": "nextPageToken,"
            + "items(snippet(title, resourceId(videoId), channelTitle))",
            "maxResults": max_results,
            "playlistId": id,
            "key": youtube_api_key,
            "pageToken": page,
        }
        logger.info("session.get triggered: list_playlistitems")
        result = cls.session.get(API.endpoint + "playlistItems", params=query)
        return result.json()



import json
import re

from mopidy_youtube import logger
from mopidy_youtube.youtube import Client, Video


class scrAPI(Client):
    """
    Indirect access to YouTube data, without API, using regex
    """

    endpoint = "https://www.youtube.com/"

    @classmethod
    def search(cls, q):
        """
        search for videos and playlists
        """

        search_results = []

        # assume 20 results per page
        pages = int(Video.search_results / 20) + (Video.search_results % 20 > 0)

        logger.info("session.get triggered: search")

        rs = [
            {
                "search_query": q.replace(" ", "+"),
                "page": page + 1,
                "app": "desktop",
                "persist_app": 1,
            }
            for page in range(pages)
        ]

        for result in [cls.run_search(r) for r in rs]:
            search_results.extend(result)

        return json.loads(
            json.dumps(
                {
                    "items": [
                        x
                        for _, x in zip(
                            range(Video.search_results), search_results
                        )
                    ]
                },
                sort_keys=False,
                indent=1,
            )
        )

    @classmethod
    def run_search(cls, query):
        result = cls.session.get(cls.endpoint + "results", params=query)
        regex = (
            r'(?:\<li\>)(?:.|\n)*?\<a href\=(["\'])\/watch\?v\=(?P<id>.{11})'
            r"(?:\&amp\;list\=(?:(?P<playlist>PL.*?)\1)?"
            r"(?:.|\n)*?class\=\1formatted-video-count-label\1\>[^\d]*"
            r"(?P<itemCount>\d*))?(?:.|\n)*?title\=\1(?P<title>.+?)\1.+?"
            r"(?:(?:Duration[^\d]+(?:(?P<durationHours>\d+)\:)?"
            r"(?P<durationMinutes>\d{1,2})\:"
            r"(?P<durationSeconds>\d{2})[^\d].*?)?)?"
            r"\<a href\=\1(?:(?:(?P<uploaderUrl>/"
            r'(?:user|channel)/[^"\']+)\1[^>]+>)?'
            r"(?P<uploader>[^<]+).+class\=\1"
            r"(?:yt-lockup-description|yt-uix-sessionlink)"
            r"[^>]*>(?P<description>.*?)\<\/div\>)?"
        )
        items = []

        for match in re.finditer(regex, result.text):
            duration = cls.format_duration(match)
            if match.group("playlist") is not None:
                item = {
                    "id": {
                        "kind": "youtube#playlist",
                        "playlistId": match.group("playlist"),
                    },
                    "contentDetails": {"itemCount": match.group("itemCount")},
                }
            else:
                item = {
                    "id": {
                        "kind": "youtube#video",
                        "videoId": match.group("id"),
                    },
                }
                if duration != "":
                    item.update(
                        {"contentDetails": {"duration": "PT" + duration}}
                    )
            item.update(
                {
                    "snippet": {
                        "title": match.group("title"),
                        # TODO: full support for thumbnails
                        "thumbnails": {
                            "default": {
                                "url": "https://i.ytimg.com/vi/"
                                + match.group("id")
                                + "/default.jpg",
                                "width": 120,
                                "height": 90,
                            },
                        },
                    },
                }
            )

            if match.group("uploader") is not None:
                item["snippet"].update(
                    {"channelTitle": match.group("uploader")}
                )
            else:
                item["snippet"].update({"channelTitle": "NA"})
            items.append(item)
        return items

    @classmethod
    def list_videos(cls, ids):
        """
        list videos
        """

        regex = (
            r'<div id="watch7-content"(?:.|\n)*?'
            r'<meta itemprop="name" content="'
            r'(?P<title>.*?)(?:">)(?:.|\n)*?'
            r'<meta itemprop="duration" content="'
            r'(?P<duration>.*?)(?:">)(?:.|\n)*?'
            r'<link itemprop="url" href="http://www.youtube.com/'
            r'(?:user|channel)/(?P<channelTitle>.*?)(?:">)(?:.|\n)*?'
            r"</div>"
        )
        items = []

        for id in ids:
            query = {"v": id, "app": "desktop", "persist_app": 1}
            logger.info("session.get triggered: list_videos")
            result = cls.session.get(cls.endpoint + "watch", params=query)
            for match in re.finditer(regex, result.text):
                item = {
                    "id": id,
                    "snippet": {
                        "title": match.group("title"),
                        "channelTitle": match.group("channelTitle"),
                    },
                    "contentDetails": {"duration": match.group("duration")},
                }
                items.append(item)
        return json.loads(
            json.dumps({"items": items}, sort_keys=False, indent=1)
        )

    @classmethod
    def list_playlists(cls, ids):
        """
        list playlists
        """

        regex = (
            r'<div id="pl-header"(?:.|\n)*?"'
            r"(?P<thumbnail>https://i\.ytimg\.com\/vi\/.{11}/).*?\.jpg"
            r'(?:(.|\n))*?(?:.|\n)*?class="pl-header-title"'
            r'(?:.|\n)*?\>\s*(?P<title>.*)(?:.|\n)*?<a href="/'
            r"(user|channel)/(?:.|\n)*? >"
            r"(?P<channelTitle>.*?)</a>(?:.|\n)*?"
            r"(?P<itemCount>\d*) videos</li>"
        )
        items = []

        for id in ids:
            query = {
                "list": id,
                "app": "desktop",
                "persist_app": 1,
            }
            logger.info("session.get triggered: list_playlists")
            result = cls.session.get(cls.endpoint + "playlist", params=query)
            for match in re.finditer(regex, result.text):
                item = {
                    "id": id,
                    "snippet": {
                        "title": match.group("title"),
                        "channelTitle": match.group("channelTitle"),
                        "thumbnails": {
                            "default": {
                                "url": match.group("thumbnail") + "default.jpg",
                                "width": 120,
                                "height": 90,
                            },
                        },
                    },
                    "contentDetails": {"itemCount": match.group("itemCount")},
                }
                items.append(item)

        return json.loads(
            json.dumps({"items": items}, sort_keys=False, indent=1)
        )

    @classmethod
    def list_playlistitems(cls, id, page, max_results):
        """
        list playlist items
        """
        query = {"list": id, "app": "desktop", "persist_app": 1}
        logger.info("session.get triggered: list_playlist_items")

        result = cls.session.get(cls.endpoint + "playlist", params=query)

        # TODO: implement pagination to get playlist entries > 100

        regex = (
            r'<tr class\=\"pl-video.*\" data-title\=\"(?P<title>.+?)".*?'
            r"<a href\=\"\/watch\?v\=(?P<id>.{11})\&amp;(?:.|\n)*?"
            r"(?P<thumbnail>https://i\.ytimg\.com\/vi\/.{11}/).*?\.jpg"
            r'(?:.|\n)*?<div class="pl-video-owner">(?:.|\n)*?'
            r"/(?:user|channel)/(?:.|\n)*? >(?P<channelTitle>.*?)</a>"
            r'(?:.|\n)*?<div class="timestamp">.*?">(?:(?:'
            r"(?P<durationHours>[0-9]+)\:)?"
            r"(?P<durationMinutes>[0-9]+)\:"
            r"(?P<durationSeconds>[0-9]{2}))"
            r"(?:.|\n)*?</div></td></tr>"
        )
        items = []

        for match in re.finditer(regex, result.text):
            duration = cls.format_duration(match)
            item = {
                "id": match.group("id"),
                "snippet": {
                    "resourceId": {"videoId": match.group("id")},
                    "title": match.group("title"),
                    "channelTitle": match.group("channelTitle"),
                    "thumbnails": {
                        "default": {
                            "url": match.group("thumbnail") + "default.jpg",
                            "width": 120,
                            "height": 90,
                        },
                    },
                },
            }
            if duration != "":
                item.update({"contentDetails": {"duration": "PT" + duration}})

            items.append(item)

        del items[max_results:]

        result = json.loads(
            json.dumps(
                {"nextPageToken": None, "items": items},  # noqa: E501
                sort_keys=False,
                indent=1,
            )
        )
        return result
